/*
 * Check whether a given user is required to use s/key or not
 *
 * This code was inspired by and is a complete rewrite of
 * Wietse Venema's skeyaccess.c, as found in his logdaemon package.
 *
 * Of course, all bugs are mine --okir@monad.swb.de
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include <syslog.h>
#include <pwd.h>

#include "skey.h"

struct login_info {
	const char *	user;
	const char *	port;
	const char *	host;
	struct in_addr	addr;
};

static int	check_access(const struct login_info *info);
static int	match(const struct login_info *info);
static int	match_host(const struct login_info *info, const char *arg);
static int	match_network(const struct login_info *info, const char *arg);
static int	get_hostname(struct login_info *, const char *);
static char *	chop(char *buffer);

int
skeyaccess(const char *logname, const char *host, const char *port)
{
	struct login_info info;

	memset(&info, 0, sizeof(info));
	info.user = logname;

	if (host) {
		int	retval;

		if ((retval = get_hostname(&info, host)) < 0)
			return retval;
		if (info.host == 0)
			info.host = host;
	}

	if (port && !strncmp(port, "/dev/", 5))
		port += 5;
#ifdef linux
	/* Translate virtual console ports */
	if (port && !strncmp(port, "tty", 3)
	 && port[3+strspn(port+3, "0123456789")] == '\0')
		port = "console";
#endif
	info.port = port;

	return check_access(&info);
}

static int
check_access(const struct login_info *info)
{
	char	buffer[1024], *word;
	FILE	*fp;
	int	line, permission, status;

	/* Implicit check: console is always okay */
	if (info->port && !strcmp(info->port, "console"))
		return SKEY_ACCESS_CONSOLE;

	if ((fp = fopen(ACCESSFILE, "r")) == NULL)
		return SKEY_ACCESS_NOFILE;

	status = line = 0;
	do {
		unsigned int	count = 0;

		/* Default to deny */
		permission = SKEY_ACCESS_SKEY_ONLY;

		while (1) {
			if (!fgets(buffer+count, sizeof(buffer)-count, fp))
				break;
			line++;

			/* Deal with backslash-newline */
			count = strlen(buffer);
			while (count && strchr(" \t\n", buffer[count-1]))
				buffer[--count] = '\0';
			if (!count || buffer[count-1] != '\\')
				break;
			buffer[count-1] = ' ';
		}

		if (count == 0 && feof(fp))
			break;

		if (!(word = chop(buffer)))
			continue;

		if (!strcmp(word, "passwd")
		 || !strcmp(word, "permit"))
			permission = SKEY_ACCESS_PASSWDOK;
		else
		if (!strcmp(word, "skey")
		 || !strcmp(word, "deny"))
			permission = SKEY_ACCESS_SKEY_ONLY;
		else {
			status = SKEY_ACCESS_SYNTAXERR;
			break;
		}

		status = match(info);
	} while (status == 0);
	fclose(fp);

	return (status < 0)? status : permission;
}

static int
strequal(const char *s1, const char *s2)
{
	return s1 && s2 && !strcasecmp(s1, s2);
}

static int
match(const struct login_info *info)
{
	char	*word, *arg;
	int	match = 1;

	while (match && (word = chop(NULL)) != NULL) {
		if (!(arg = chop(NULL)))
			return SKEY_ACCESS_SYNTAXERR;

		if (!strcmp(word, "port")) {
			match = strequal(arg, info->port);
		} else
		if (!strcmp(word, "user")) {
			match = strequal(arg, info->user);
		} else
		if (!strcmp(word, "hostname") || !strcmp(word, "host")) {
			match = match_host(info, arg);
		} else
		if (!strcmp(word, "internet") || !strcmp(word, "net")) {
			match = match_network(info, arg);
		} else
			return SKEY_ACCESS_SYNTAXERR;
	}

	return match;
}

static int
match_host(const struct login_info *info, const char *arg)
{
	struct in_addr	addr;

	if (info->host == NULL)
		return 0;
	if (inet_aton(arg, &addr))
		return addr.s_addr == info->addr.s_addr;
	return strequal(info->host, arg);
}

static int
match_network(const struct login_info *info, const char *arg)
{
	struct in_addr	addr, mask;

	if (info->host == NULL)
		return 0;
	if (!inet_aton(arg, &addr))
		return SKEY_ACCESS_SYNTAXERR;
	if (!(arg = chop(NULL)) || !inet_aton(arg, &mask))
		return SKEY_ACCESS_SYNTAXERR;
	/* XXX: make sure mask is really a netmask, i.e 1...10..0 */
	return ((info->addr.s_addr & mask.s_addr) == addr.s_addr);
}

static int
get_hostname(struct login_info *info, const char *hostname)
{
	static char	buffer[1024];
	struct hostent	*hp;
	struct in_addr	addr;
	unsigned int	okay = 0;

	if (strlen(hostname) >= sizeof(buffer))
		return SKEY_ACCESS_EINVAL;
	strcpy(buffer, hostname);

	if (inet_aton(hostname, &addr) != 0) {
		char	**addrs;

		hp = gethostbyaddr((char *) &addr, sizeof(addr), AF_INET);
		if (hp == NULL)
			return 0;

		/* Copy FQDN */
		if (strlen(hp->h_name) >= sizeof(buffer))
			return SKEY_ACCESS_EINVAL;
		strcpy(buffer, hp->h_name);

		/* Forward check: verify addresses */
		hp = gethostbyname(buffer);
		if (hp == NULL)
			return 0;

		addrs = hp->h_addr_list;
		while (*addrs && !okay)
			okay = !memcmp(*addrs++, &addr, sizeof(addr));
	} else {
		char	**aliases;

		hp = gethostbyname(hostname);
		if (hp == NULL)
			return 0;

		/* Copy FQDN */
		if (strlen(hp->h_name) >= sizeof(buffer))
			return SKEY_ACCESS_EINVAL;
		strcpy(buffer, hp->h_name);

		/* Backward check */
		hp = gethostbyaddr((char *) hp->h_addr, sizeof(addr), AF_INET);
		if (hp == NULL)
			return 0;

		okay = !strcasecmp(hp->h_name, buffer);
		aliases = hp->h_aliases;
		while (*aliases && !okay)
			okay = !strcmp(*aliases++, buffer);
	}

	if (okay)
		info->host = buffer;
	return okay;
}

static char *
chop(char *buffer)
{
	char	*word;

	word = strtok(buffer, " #\t\n");
	if (word && !*word)
		word = NULL;
	return word;
}
