/*
 * PAM S/Key module
 *
 * Copyright (C) 1999, Olaf Kirch <okir@caldera.de>
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/types.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <ctype.h>

#include <security/_pam_macros.h>
#include <security/pam_modules.h>

#ifndef LINUX_PAM
#include <security/pam_appl.h>
#endif  /* LINUX_PAM */

#include "skey.h"

enum {
	_SKEY_DEBUG	= 0x0001,
	_SKEY_ECHO_OFF	= 0x0002,
	_SKEY_CHECK	= 0x0004,
	_SKEY_PASSWDOK	= 0x0008,
	_SKEY_SILENT	= 0x0020,
	_SKEY_NONULL	= 0x0040,
};
#define _SKEY_DEFAULT	(_SKEY_ECHO_OFF)

#define _SKEY_INFO_MSG		"S/Key one time passwords are in effect."
#define _SKEY_CHALLENGE_MSG	"Challenge: %s"
#define _SKEY_PROMPT_MSG	"Please enter s/key response: "
#define _SKEY_PROMPT_PASSWD_MSG	"Please enter s/key response or plaintext password: "

static int	_skey_checkaccess(pam_handle_t *, int, const char *);
static int	_skey_getresponse(pam_handle_t *pamh, int ctrl,
			const char *challenge, char **responsep);
static void	_skey_log(int lvl, const char *fmt, ...);
static int	_skey_setctrl(int flags, int argc, const char **argv);
static int	_skey_converse(pam_handle_t *, int, int,
			const struct pam_message **, struct pam_response **);

PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	const char	*logname;
	struct skey	skey;
	char		challenge[64], *response;
	int		ctrl, retval;

	ctrl = _skey_setctrl(flags, argc, argv);
	retval = pam_get_user(pamh, &logname, "login:");

	if (retval != PAM_SUCCESS) {
		if (retval == PAM_CONV_AGAIN) {
			D(("get user: waiting for data"));
			return PAM_INCOMPLETE;
		}
		if (ctrl & _SKEY_DEBUG)
			_skey_log(LOG_DEBUG, "auth could not get user name");
		return retval;
	}
	if (logname == NULL) {
		D(("null user name"));
		return PAM_USER_UNKNOWN;
	}
	if (!isalnum(*logname)) {
		D(("bad user name"));
		_skey_log(LOG_NOTICE, "bad user name [%s]", logname);
		return PAM_USER_UNKNOWN;
	}
	if (ctrl & _SKEY_DEBUG)
		_skey_log(LOG_DEBUG, "got user name [%s]", logname);

	if (ctrl & _SKEY_CHECK) {
		D(("checking access"));
		retval = _skey_checkaccess(pamh, ctrl, logname);
		if (retval == SKEY_ACCESS_NOFILE
		 || retval == SKEY_ACCESS_CONSOLE)
			return PAM_IGNORE;
		if (retval < 0)
			return PAM_SESSION_ERR;
		if (retval == 1)
			ctrl |= _SKEY_PASSWDOK;
	}

	retval = skeyinfo(&skey, logname, challenge);
	if (retval < 0) {
		_skey_log(LOG_NOTICE, "s/key database error");
		return PAM_ABORT;
	}
	if (retval == 0) {
		D(("no s/key for %s", logname));
		if (ctrl & _SKEY_DEBUG)
			_skey_log(LOG_DEBUG, "no s/key for user %s", logname);
		if (ctrl & _SKEY_NONULL)
			return PAM_AUTH_ERR;
		return PAM_USER_UNKNOWN;
	}
	if (ctrl & _SKEY_DEBUG)
		_skey_log(LOG_DEBUG, "got challenge %s", challenge);

	/* Okay, we have a challenge */
	retval = _skey_getresponse(pamh, ctrl, challenge, &response);
	if (retval != PAM_SUCCESS) {
		if (retval != PAM_INCOMPLETE && (ctrl & _SKEY_DEBUG))
			_skey_log(LOG_DEBUG, "auth could not get response");
		if (response)
			_pam_drop(response);
		return retval;
	}

	/* Empty response means don't try s/key.
	 * (and don't store token so subsequent modules do their
	 * own prompting) */
	if (response[0] == '\0')
		return PAM_IGNORE;

	/* Store auth token */
	retval = pam_set_item(pamh, PAM_AUTHTOK, response);
	if (retval != PAM_SUCCESS) {
		_skey_log(LOG_NOTICE, "unable to save auth token");
		return retval;
	}

	/* Now try to verify response */
	retval = skeyverify(&skey, response);
	_pam_drop(response);

	if (retval < 0) {
		if (ctrl & _SKEY_DEBUG)
			_skey_log(LOG_DEBUG, "bad response for %s", logname);
		if (ctrl & _SKEY_PASSWDOK)
			_skey_log(LOG_DEBUG, "(continuing)");
		if (ctrl & _SKEY_PASSWDOK)
			return PAM_IGNORE;
		return PAM_AUTH_ERR;
	}

	return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	return PAM_IGNORE;
}

static int
_skey_setctrl(int flags, int argc, const char **argv)
{
	int	ctrl;

	D(("called."));

	ctrl = _SKEY_DEFAULT;
	if (flags & PAM_DISALLOW_NULL_AUTHTOK)
		ctrl |= _SKEY_NONULL;
	if (flags & PAM_SILENT)
		ctrl |= _SKEY_SILENT;

	while (argc--) {
		if (!strcmp(argv[argc], "debug"))
			ctrl |= _SKEY_DEBUG;
		else if (!strcmp(argv[argc], "echo=off"))
			ctrl |= _SKEY_ECHO_OFF;
		else if (!strcmp(argv[argc], "echo=on"))
			ctrl &= ~_SKEY_ECHO_OFF;
		else if (!strcmp(argv[argc], "checkaccess"))
			ctrl |= _SKEY_CHECK;
		else
			_skey_log(LOG_ERR, "unrecognized option %s",
				argv[argc]);
	}

	return ctrl;
}

static int
_skey_checkaccess(pam_handle_t *pamh, int ctrl, const char *logname)
{
	char	*host, *port;
	int	retval;

	if (pam_get_item(pamh, PAM_RHOST, (const void **) &host) != PAM_SUCCESS)
		host = 0;
	if (pam_get_item(pamh, PAM_TTY, (const void **) &port) != PAM_SUCCESS)
		port = 0;

	if (ctrl & _SKEY_DEBUG)
		_skey_log(LOG_DEBUG,
			"Calling skeyaccess(%s, %s, %s)", logname, host, port);
	retval = skeyaccess(logname, host, port);
	if (retval < 0)
		_skey_log(LOG_NOTICE,
			"Error in S/Key access check, status %d\n", retval);
	return retval;
}

static inline struct pam_message *
_skey_addmessage(struct pam_message *p, int style, const char *msg)
{
	p->msg_style = style;
	p->msg = msg;
	return p;
}

static int
_skey_getresponse(pam_handle_t *pamh, int ctrl,
		  const char *challenge, char **responsep)
{
	struct pam_message	msg[3];
	const struct pam_message *m[3];
	struct pam_response	*resp = NULL;
	const char		*prompt;
	char			buffer[256];
	unsigned int		i = 0;
	int			pstyle, retry = 0, retval;

	/* Initialize */
	*responsep = NULL;

	/* Check whether we've previously prompted */
	retval = pam_get_data(pamh, "skey_retry", (const void **) &retry);
	if (retval != PAM_SUCCESS)
		retry = 0;

	/* Display prompts only if not retrying */
	if (!retry) {
		if (!(ctrl & _SKEY_SILENT)) {
			m[i] = _skey_addmessage(msg+i, PAM_TEXT_INFO,
						_SKEY_INFO_MSG);
			i++;
		}

		snprintf(buffer, sizeof(buffer),
				_SKEY_CHALLENGE_MSG, challenge);
		m[i] = _skey_addmessage(msg+i, PAM_TEXT_INFO, buffer);
		i++;
	}
	retry++;

	if (ctrl & _SKEY_ECHO_OFF) {
		pstyle = PAM_PROMPT_ECHO_OFF;
	} else {
		pstyle = PAM_PROMPT_ECHO_ON;
	}
	prompt = _SKEY_PROMPT_MSG;
	if (ctrl & _SKEY_PASSWDOK)
		prompt = _SKEY_PROMPT_PASSWD_MSG;
	m[i] = _skey_addmessage(msg+i, pstyle, prompt);
	i++;

	/* Do the conversation */
	retval = _skey_converse(pamh, ctrl, i, m, &resp);

	/* Save retry state */
	pam_set_data(pamh, "skey_retry", (void **) retry, NULL);

	if (retval != PAM_SUCCESS) {
		if (retval == PAM_CONV_AGAIN) {
			D(("get user: waiting for data"));
			/* record that we've already displayed the
			 * prompts */
			retval = PAM_INCOMPLETE;
		}
	} else if (resp != NULL) {
		*responsep = resp[i-1].resp;
		resp[i-1].resp = NULL;
	} else {
		retval = PAM_AUTHTOK_RECOVER_ERR;
	} 

	if (resp)
		_pam_drop_reply(resp, i);

	/* Save retry state */
	retry = 0;
	pam_set_data(pamh, "skey_retry", (void **) retry, NULL);

	return retval;
}

/*
 * Another of those functions that are repeated in every source file,
 * like a mantra.
 */
static int
_skey_converse(pam_handle_t *pamh, int ctrl, int count,
		const struct pam_message **msg, struct pam_response **resp)
{
	struct pam_conv	*conv;
	int		retval;

	retval = pam_get_item(pamh, PAM_CONV, (const void **) &conv);
	if (retval == PAM_SUCCESS)
		retval = conv->conv(count, msg, resp, conv->appdata_ptr);
	return retval;
}

static void
_skey_log(int lvl, const char *fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	closelog();
	openlog("pam_skey", LOG_CONS|LOG_PID, LOG_AUTH);
	vsyslog(lvl, fmt, ap);
	va_end(ap);
	closelog();
}
