/* Stand-alone program for computing responses to S/Key challenges.
 * Takes the iteration count and seed as command line args, prompts
 * for the user's key, and produces both word and hex format responses.
 *
 * Usage example:
 *	>skey 88 ka9q2
 *	Enter password:
 *	OMEN US HORN OMIT BACK AHOY
 *	C848 666B 6435 0A93
 *	>
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>

#include "skey.h"

static int	opt_print = 0,
		opt_hex = 0;

static void	usage(int exval);
static void	printkey(const unsigned char *);

int
main(int argc, char **argv)
{
	unsigned int	i, sequence = 0, count;
	const char	*seed = 0;
	char		*seqstr, *passwd;
	unsigned char	key[8];
	struct skey	skey;

	if (argc >= 2 && !strcmp(argv[1], "init")) {
		argv[1] = "keyinit";
		execv(KEYINIT_PATH, argv+1);
		perror(KEYINIT_PATH);
		exit(1);
	}

	count = 1;
	while ((i = getopt(argc, argv, "hn:x")) != EOF) {
		switch (i) {
		case 'h':
			usage(0);
		case 'n':
			count = atoi(optarg);
			opt_print = 1;
			break;
		case 'x':
			opt_hex = 1;
			break;
		default:
			usage(1);
		}
	}

	/* May be preceded with s/key from challenge string */
	if (optind < argc && !strcmp(argv[optind], "s/key"))
		optind++;

	/* Remaining args can be either number/seed or `number seed' */
	if (opt_print && optind == argc) {
		/* Home base mode: print out several OTPs based on 
		 * local key file */
		const char	*logname;

		if ((logname = getenv("LOGNAME")) == NULL
		 && (logname = getenv("USER")) == NULL) {
			fprintf(stderr, "Unable to determine user name.\n");
			exit(1);
		}
		if (skeyinfo(&skey, logname, NULL) < 0) {
			fprintf(stderr, "No key for %s\n", logname);
			exit(1);
		}
		seed = skey.seed;
		sequence = skey.n;
		seqstr = NULL;
	} else if (optind == argc - 1) {
		/* Argument is number/seed */
		char	*sp;

		seqstr = argv[optind++];
		if ((sp = strchr(seqstr, '/')) == NULL)
			usage(1);
		*sp++ = '\0';
		seed = sp;
	} else if (optind == argc - 2) {
		/* Two args, separate number/seed */
		seqstr = argv[optind++];
		seed = argv[optind++];
	} else {
		usage(1);
	}

	/* Make sure it's a number */
	if (seqstr) {
		sequence = strtoul(seqstr, &seqstr, 10);
		if (*seqstr) {
			fprintf(stderr, "Not a number: %s\n", seqstr);
			exit(1);
		}
	}

	if (count > sequence) {
		fprintf(stderr, "Count %d greater than sequence.\n", count);
		exit(1);
	}

	fprintf(stderr,
	"Note: Do not use this program while logged in over an insecure connection.\n");

	/* Get user's secret password */
	passwd = getpass("Enter secret password: ");

	/* Crunch seed and password into starting key */
	if (keycrunch(key, seed, passwd) != 0) {
		fprintf(stderr, "key crunch failed\n");
		exit(1);
	}
	memset(passwd, 0, strlen(passwd));

	for (i = 0; i <= sequence - count; i++)
		f(key);

	if (count == 1) {
		printkey(key);
	} else {
		for (; i <= sequence; i++) {
			printf("%4d: ", i);
			printkey(key);
			f(key);
		}
	}

	return 0;
}

static void
usage(int exval)
{
	fprintf(stderr,
		"Usage: skey [-n count] <sequence #>[/] <key> \n");
	exit(exval);
}

static void
printkey(const unsigned char *key)
{
	char	buf[64];

	printf("%-29s", btoe(buf, key));
	if (opt_hex)
		printf(" %s", put8(buf, key));
	fputs("\n", stdout);
}
