/*   change password or add user to S/KEY authentication system.
 *   S/KEY is a tradmark of Bellcore
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <getopt.h>
#include <stdarg.h>

#include "skey.h"

static int	opt_ask = 0,
		opt_secure = 0,
		opt_needauth = 0,
		opt_debug = 0;

static void		update(const char *logname);
static unsigned char *	makekey(struct skey *, const char *, unsigned char *);
static int		verifykey(struct skey *, const unsigned char *);
static void		printkey(unsigned int n, const unsigned char *key);
static char *		makepassword(void);
static char *		makeseed(const char *oldseed, char *seed);
static unsigned int	makesequence(void);
static char *		readstring(const char *fmt, ...);
static void		usage(int exval);


int
main(int argc, char **argv)
{
	struct passwd	*pw = NULL;
	int		c;

	setlinebuf(stdout);
	setlinebuf(stderr);

	while ((c = getopt(argc, argv, "adhs")) != -1) {
		switch (c) {
		case 'a':
			opt_ask = 1;
			break;
		case 'd':
			opt_debug++;
			break;
		case 'h':
			usage(0);
		case 's':
			opt_secure = 1;
			break;
		default:
			usage(1);
		}
	}

	if (optind == argc - 1) {
		if (!(pw = getpwnam(argv[optind]))) {
			fprintf(stderr, "Unknown user: %s.\n", argv[optind]);
			exit(1);
		}
		if (pw->pw_uid != getuid() && getuid()) {
			fprintf(stderr, "Only root can do that.\n");
			exit(1);
		}
	} else if (optind == argc) {
		if (!(pw = getpwuid(getuid()))) {
			fprintf(stderr, "You do not exist.\n");
			exit(1);
		}
	} else {
		usage(1);
	}

	update(pw->pw_name);
	return 0;
}

static void
update(const char *logname)
{
	struct skey	skey, okey;
	char		seed[32], challenge[128], *passwd = 0;
	unsigned char	key[8];
	int		res;

	if ((res = skeyinfo(&skey, logname, challenge)) < 0) {
		fprintf(stderr, "Unable to open database.\n");
		exit(1);
	}

#ifdef AUTHENTICATE_UPDATE
	opt_needauth = (getuid() != 0);
#endif

	if (res == 1) {
		printf("Updating key for user %s\n", logname);
		printf("Old key: %s ", skey.seed);
		printkey(skey.n, skey.key);
		okey = skey;
	} else
	if (!opt_needauth) {
		printf("Adding key for user %s\n", logname);
		memset(&skey, 0, sizeof(skey));
		skey.logname = logname;
	} else {
		fprintf(stderr,
			"There is no key for user %s in the database,\n"
			"and only root is allowed to add new keys.\n"
			"Sorry.\n", logname);
		exit(1);
	}

	/* Obtain new seed and sequence */
	skey.seed = makeseed(skey.seed, seed);
	skey.n    = makesequence();

	if (opt_secure) {
		char	*resp;
		int	i = 0;

		printf("\nPlease compute the NEW hashed key on your computer\n"
		       "using the following S/KEY challenge:\n"
		       "s/key %d %s\n", skey.n, skey.seed);

		while (1) {
			if (++i > 3) {
				fprintf(stderr, "Sorry.\n");
				exit(1);
			}
			resp = readstring("Please enter response: ");
			if (etob(key, resp) == 1 || atob8(key, resp) == 0)
				break;
			printf("Invalid format, please try again.\n");
		}

	} else {
		/* Get new password and sequence, and crunch it into key */
		passwd = makepassword();
		makekey(&skey, passwd, key);
	}
	skey.key = key;

	if (!opt_needauth) {
		/* Set new key without asking for authentication */
		res = skeysetkey(&skey, NULL);
	} else if (opt_secure) {
		unsigned char	key[8];
		char		*resp;

		printf("\nBefore accepting the updated key, please\n"
		       "authenticate yourself using your OLD S/Key.\n"
		       "The challenge is %s\n", challenge);
		resp = readstring("Please enter response: ");
		if (etob(key, resp) != 1 && atob8(key, resp) < 0) {
			fprintf(stderr, "Cannot parse input.\n");
			exit(1);
		}
		res = skeysetkey(&skey, key);
	} else {
		unsigned char	key[8];

		/* First try the new password */
		res = -1;
		okey.n--;
		if (passwd && verifykey(&okey, makekey(&okey, passwd, key))) {
			res = skeysetkey(&skey, key);
		} else {
			printf("\nBefore accepting the updated key, please\n"
			       "authenticate yourself using your OLD password.\n");
			makekey(&okey, NULL, key);
			res = skeysetkey(&skey, key);
		}
	}

	if (res < 0) {
		fprintf(stderr, "Unable to update key.\n");
		exit(1);
	}
	printf("Successfully updated s/key key.\n");
}

static void
printkey(unsigned int n, const unsigned char *key)
{
	printf("[%2d] %02x%02x%02x%02x%02x%02x%02x%02x\n", n,
		key[0], key[1], key[2], key[3],
		key[4], key[5], key[6], key[7]);
}

static int
verifykey(struct skey *mp, const unsigned char *key)
{
	unsigned char	temp[8];

	memcpy(temp, key, 8);
	f(temp);
	return !memcmp(mp->key, temp, 8);
}

static unsigned char *
makekey(struct skey *mp, const char *passwd, unsigned char *key)
{
	unsigned int	n;
	char		*p = NULL;

	if (passwd == NULL)
		passwd = p = getpass("Please enter password: ");

	/* Crunch seed and password into starting key */
	if (keycrunch(key, mp->seed, passwd) < 0) {
		fprintf(stderr, "Error, key crunch failed.\n");
		exit(1);
	}

	/* Clear password if gotten temporarily */
	if (p)
		memset(p, 0, strlen(p));

	/* Apply the hash */
	for (n = 0; n <= mp->n-1; n++)
		f(key);

	return key;
}

static char *
makeseed(const char *oldseed, char *seed)
{
	char	hostname[256], *resp, *sp;
	int	l;

	/* `Increment' the default seed. This turns
	 * ab1239 into ab1240.
	 * Also does ab12zz99 -> ab13zz00
	 */
	seed[0] = '\0';
	if (oldseed) {
		l = strlen(oldseed);
		strcpy(seed, oldseed);
		while (1) {
			if (l-- == 0) {
				seed[0] = '\0';
				break;
			}
			if (seed[l] == '9') {
				seed[l] = '0';
			} else if (isdigit(seed[l])) {
				seed[l] += 1;
				break;
			}
		}
	}

	if (seed[0] == '\0') {
		gethostname(hostname, sizeof(hostname));
		if ((sp = strchr(hostname, '.')) != NULL)
			*sp = '\0';
		if (strlen(hostname) > 7)
			hostname[7] = '\0';
		strcpy(seed, hostname);
		get_random_digits(seed + strlen(seed), 5);
	}

	/* Just use the default seed without asking? */
	if (!opt_secure && !opt_ask)
		return seed;

	resp = readstring("\nEnter new seed [%s]: ", seed);
	if (*resp != '\0') {
		strncpy(seed, resp, 16);
		seed[16] = '\0';
	}

	return seed;
}

static unsigned int
makesequence(void)
{
	char		*resp, *sp;
	unsigned int	i, n;

	printf("\n"
"You can now set the sequence counts (i.e. the number of valid uses of this\n"
"password). The number must be between 1 and 9999.\n"
"\n");

	for (i = 0; ; i++) {
		if (i >= 2)
			exit(1);
		resp = readstring("Please enter number [99]: ");
		if (*resp == '\0')
			return 99;
		n = strtoul(resp, &sp, 10);
		if (*sp)
			continue;
		if (n > 0 && n < 10000)
			break;
		printf("Count must be in the range 1 to 9999\n");
	}

	return n;
}

static char *
makepassword(void)
{
	char	buffer[128], *pass;
	int	i = 0;

	printf("\n"
"Reminder: Do NOT use this method if logged in over an insecure connection!\n"
"          Use skeyinit -s instead.\n"
"\n");

	while (1) {
		if (++i > 2)
			exit(1);
		pass = getpass("Please enter your new password: ");
		if (strlen(pass) >= sizeof(buffer)) {
			fprintf(stderr, "Password too long.\n");
			continue;
		}
		strcpy(buffer, pass);
		pass = getpass("Please type password again: ");

		if (strcmp(buffer, pass)) {
			fprintf(stderr, "The two passwords do not match.\n");
			continue;
		}
		if (strlen(pass) < 4) {
			fprintf(stderr, "The password is too short.\n");
			continue;
		}
		printf("The two passwords match.\n");
		break;
	}

	memset(buffer, 0, sizeof(buffer));
	return pass;
}

static void
usage(int exval)
{
	fprintf(stderr,
		"usage: skeyinit [-s] [username]\n");
	exit(exval);
}

static char *
readstring(const char *fmt, ...)
{
	static char	buffer[256];
	char		*sp;
	va_list		ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	fflush(stdout);
	fgets(buffer, sizeof(buffer), stdin);

	if ((sp = strchr(buffer, '\n')) != NULL)
		*sp = '\0';
	return buffer;
}

#if 0
#define NAMELEN 2
int
xmain(int argc, char **argv)
{
	struct skey skey;
	int rval,n,nn,i,defaultsetup;
	char seed[18],tmp[80],key[8];
	struct passwd *ppuser,*pp;
	char defaultseed[BUFSIZ], passwd[256],passwd2[256] ;
	int ch;


	time_t now;
	struct tm *tm;
	char tbuf[27],buf[60];
	char lastc, me[80];
	int l;

	time(&now);
#if 0	/* Choose a more random seed */
	tm = localtime(&now);
	strftime(tbuf, sizeof(tbuf), "%M%j", tm);
#else
	sprintf(tbuf, "%05ld", (long) (now % 100000));
#endif
	gethostname(defaultseed,sizeof(defaultseed));
	strcpy(&defaultseed[NAMELEN],tbuf);

	pp = ppuser = getpwuid(getuid());
	if(pp == NULL){
		printf("User unknown\n");
		return 1;
	}
	strcpy(me,pp->pw_name);
	defaultsetup = 1;

	/*
	 * Parse JCL.
	 */
	while ((ch = getopt(argc, argv, "s")) > 0) {
		switch (ch) {
		case 's':
			defaultsetup = 0;
			break;
		default:
			fprintf(stderr, "Usage: %s [-s] [username]\n", argv[0]);
			exit(1);
		}
	}
	if (argc > optind + 1) {
		fprintf(stderr, "Usage: %s [-s] [username]\n", argv[0]);
		exit(1);
	}
	if (argc == optind + 1) {
		pp = getpwnam(argv[optind]);
		if (pp == NULL){
			printf("User unknown\n");
			return 1;
		}
	}

	if(strcmp( pp->pw_name,me) != 0){
		if(getuid() != 0){
			/* Only root can change other's passwds */
			printf("Permission denied.\n");
			return(1);
		}
	}



	rval = skeylookup(&skey,pp->pw_name);
	switch(rval){
	case -1:
		perror("error in opening database");
		return 1;
	case 0:
		printf("Updating %s:\n",pp->pw_name);
		printf("Old key: %s\n",skey.seed);
		/* lets be nice if they have a skey.seed that ends in 0-8 just add one*/
		l = strlen(skey.seed);
		if( l > 0){
			lastc = skey.seed[l-1];
			if( isdigit(lastc) && lastc != '9' ){
				strcpy(defaultseed, skey.seed);
				defaultseed[l-1] = lastc + 1;
			}
			if( isdigit(lastc) && lastc == '9' && l < 16){
				strcpy(defaultseed, skey.seed);
				defaultseed[l-1] = '0';
				defaultseed[l] = '0';
				defaultseed[l+1] = '\0';
			}
		}
		break;
	case 1:
		skey.val = 0;	/* XXX */
		printf("Adding %s:\n",pp->pw_name);
		break;
	}
    n = 99;
    if( ! defaultsetup){
	printf("Reminder you need the 6 english words from the skey command.\n");
	for(i=0;;i++){
		if(i >= 2) exit(1);
		printf("Enter sequence count from 1 to 9999: ");
		fgets(tmp,sizeof(tmp),stdin);
		n = atoi(tmp);
		if(n > 0 && n < 10000)
			break;	/* Valid range */
		printf("Count must be > 0 and < 10000\n");
	}
    }
    if( !defaultsetup){
	printf("Enter new key [default %s]: ", defaultseed);
	fflush(stdout);
	fgets(seed,sizeof(seed),stdin);
	rip(seed);
	if(strlen(seed) > 16){
		printf("Seed truncated to 16 chars\n");
		seed[16] = '\0';
	}
	if( seed[0] == '\0') strcpy(seed,defaultseed);
	for(i=0;;i++){
		if(i >= 2) exit(1);
		printf("s/key %d %s\ns/key access password: ",n,seed);
		fgets(tmp,sizeof(tmp),stdin);
		rip(tmp);
		if(tmp[0] == '?'){
			printf("Enter 6 English words from secure S/Key calculation.\n");
			continue;
		}
		if(tmp[0] == '\0'){
			exit(1);
		}
		if(etob(key,tmp) == 1 || atob8(key,tmp) == 0)
			break;	/* Valid format */
		printf("Invalid format, try again with 6 English words.\n");
	}
    } else {
	/* Get user's secret password */
	fprintf(stderr,"Reminder - Only use this method if you are directly connected.\n");
	fprintf(stderr,"If you are using telnet or rlogin exit with no password and use keyinit -s.\n");
	for(i=0;;i++){
		if(i >= 2) exit(1);
		fprintf(stderr,"Enter secret password: ");
		readpass(passwd,sizeof(passwd));
		if(passwd[0] == '\0'){
			exit(1);
		}
		fprintf(stderr,"Again secret password: ");
		readpass(passwd2,sizeof(passwd));
		if(passwd2[0] == '\0'){
			exit(1);
		}
		if(strlen(passwd) < 4 && strlen(passwd2) < 4) {
			fprintf(stderr, "Sorry your password must be longer\n\r");
			exit(1);
		}
		if(strcmp(passwd,passwd2) == 0) break;
		fprintf(stderr, "Sorry no match\n");
		
	
	}
	strcpy(seed,defaultseed);

	/* Crunch seed and password into starting key */
	if(keycrunch(key,seed,passwd) != 0){
		fprintf(stderr,"%s: key crunch failed\n",argv[0]);
		return 1;
	}
	nn = n;
	while(nn-- != 0)
		f(key);
    }
	time(&now);
	tm = localtime(&now);
	strftime(tbuf, sizeof(tbuf), " %b %d,%Y %T", tm);
        if (skey.val == NULL)
                  skey.val = (char *) malloc(16+1);


	btoa8(skey.val,key);
	fprintf(skey.keyfile,"%s %04d %-16s %s %-21s\n",pp->pw_name,n,
		seed,skey.val, tbuf);
	fclose(skey.keyfile);
	printf("\nID %s s/key is %d %s\n",pp->pw_name,n,seed);
	printf("%s\n",btoe(buf,key));
#ifdef HEXIN
	printf("%s\n",put8(buf,key));
#endif
	return 0;
}
#endif
