/*
 *   Login code for S/KEY Authentication.
 *   S/KEY is a trademark of Bellcore.
 *
 *   Mink is the former name of the S/KEY authentication system.
 *   Many references for mink  may still be found in this program.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <errno.h>
#include <syslog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include "skey.h"

static int	skey_update(const struct skey *mp, const char *key);
static int	skey_open(struct skey *mp, int lock);
static int	skey_close(struct skey *mp, int commit);
static int	skey_copyperms(int lockfd);
static int	skey_getline(struct skey *mp, char *, unsigned int);
static void	skey_putline(struct skey *mp);

/* Look up skey info for user 'name'. If successful, fill in the caller's
 * skey structure and return 0. If unsuccessful (e.g., if name is unknown)
 * return -1. If an optional challenge string buffer is given, update it.
 *
 * The file read/write pointer is left at the start of the
 * record.
 */
int
skeyinfo(struct skey *mp, const char *name, char *ss)
{
	int	ret;

	if ((ret = skeylookup(mp, name)) <= 0)
		return ret;

	if (ss != 0)
		sprintf(ss, "s/key %d %s", mp->n - 1, mp->seed);
	return 1;
}	

/*
 * Return the skey challenge for the user specified in struct skey.
 */
int
skeychallenge(struct skey *mp, char *ss)
{
	sprintf(ss, "s/key %d %s", mp->n - 1, mp->seed);
	return 0;
}	

/*
 * Find an entry in the key file.
 */
int
skeylookup(struct skey *mp, const char *name)
{
	char	buffer[SKEY_BUFSIZE];
	int	found;

	if (skey_open(mp, 0) < 0)
		return -1;

	/* Look up user name in database */
	found = 0;
	while (!feof(mp->keyfile)) {
		if (skey_getline(mp, buffer, sizeof(buffer)) == 0
		 && !strcmp(mp->logname, name)) {
			found = 1;
			break;
		}
	}
	skey_close(mp, 0);
	return found;
}

/*
 * Verify response to a s/key challenge, and
 * update the database.
 */
int
skeyverify(struct skey *mp, const char *response)
{
	char            key[8];

	/* Convert response to binary */
	if (response == NULL
	 || (etob(key, response) != 1 && atob8(key, response) < 0))
		return -1;

	mp->key = key;
	mp->n   = mp->n - 1;
	return skey_update(mp, key);
}

/*
 * Set a new skey, optionally authenticating with the previous key
 * All new key information is contained in the skey struct passed by
 * the caller.
 */
int
skeysetkey(struct skey *mp, const unsigned char *key)
{
	return skey_update(mp, key);
}

/*
 * Update the key file, matching the old key against what's currently
 * in the database.
 */
static int
skey_update(const struct skey *mp, const char *key)
{
	struct skey	skey;
	int		match;

	/* Open and lock the file for write update */
	if (skey_open(&skey, 1) < 0)
		return -1;

	match = 0;
	while (!feof(skey.keyfile)) {
		char	buffer[SKEY_BUFSIZE];

		if (skey_getline(&skey, buffer, sizeof(buffer))
		 || strcmp(skey.logname, mp->logname)) {
			/* Just copy line */
			fputs(buffer, skey.lockfile);
			continue;
		}

		/* A key is given; hash it and compare it to what's in
		 * the database.
		 */
		if (key != NULL) {
			unsigned char	hashkey[8];

			/* Compute hashkey = f(key) */
			memcpy(hashkey, key, 8);
			f(hashkey);

			if (memcmp(skey.key, hashkey, 8) != 0) {
				/* Wrong response */
				syslog(LOG_NOTICE,
					"bad s/key response for %s",
					mp->logname);
				syslog(LOG_NOTICE,
					"Wanted %02x%02x%02x%02x%02x%02x%02x%02x",
					skey.key[0], skey.key[1],
					skey.key[2], skey.key[3],
					skey.key[4], skey.key[5],
					skey.key[6], skey.key[7]);
				syslog(LOG_NOTICE,
					"Got    %02x%02x%02x%02x%02x%02x%02x%02x",
					hashkey[0], hashkey[1],
					hashkey[2], hashkey[3],
					hashkey[4], hashkey[5],
					hashkey[6], hashkey[7]);
				skey_close(&skey, 0);
				return -1;
			}
		}
		/* Everything's dandy. Update the file */
		skey.seed = mp->seed;
		skey.key  = mp->key;
		skey.n    = mp->n;
		skey_putline(&skey);
		match = 1;
	}

	/* We're supposed to create a new entry: */
	if (key == NULL && !match) {
		skey.logname = mp->logname;
		skey.seed    = mp->seed;
		skey.key     = mp->key;
		skey.n       = mp->n;
		skey_putline(&skey);
		match = 1;
	}

	/* Close files, and commit change if match was found */
	if (skey_close(&skey, match) < 0)
		match = 0;

	if (match)
		syslog(LOG_NOTICE,
		       "Updated s/key %s %d/%s", mp->logname, mp->n, mp->seed);

	return match? 0 : -1;
}

/*
 * Open the key file
 */
static int
skey_open(struct skey *mp, int lock)
{
	/* Clear all state */
	memset(mp, 0, sizeof(*mp));

	/* Try to open file */
	mp->keyfile = fopen(KEYFILE, "r");

	/* If we're supposed to lock the file, create it */
	if (mp->keyfile == 0 && lock) {
		int	fd;

		fd = open(KEYFILE, O_WRONLY|O_CREAT|O_EXCL, 0600);
		if (fd < 0) {
			syslog(LOG_NOTICE, "create %s: %m", KEYFILE);
			return -1;
		}
		mp->keyfile = fdopen(fd, "r");
	}
	if (mp->keyfile == NULL) {
		syslog(LOG_NOTICE, "open %s: %m", KEYFILE);
		return -1;
	}

	/* Now lock the file */
	if (lock) {
		char	lockname[256];
		int	retries = 0, fd;

		sprintf(lockname, "%s.lock", KEYFILE);
		while (1) {
			fd = open(lockname, O_WRONLY|O_CREAT|O_EXCL, 0644);
			if (fd >= 0) {
				/* Copy permissions */
				if (skey_copyperms(fd) >= 0)
					mp->lockfile = fdopen(fd, "w");
				if (mp->lockfile == NULL) {
					int	oerrno = errno;

					unlink(lockname);
					close(fd);
					errno = oerrno;
				}
				break;
			}
			if (++retries > 5 || errno != EEXIST)
				break;
			sleep(1);
		}
		if (mp->lockfile == NULL) {
			syslog(LOG_ERR, "cannot create lock %s: %m", lockname);
			skey_close(mp, 0);
			return -1;
		}
	}

	return 0;
}

/*
 * Close skey file
 */
static int
skey_close(struct skey *mp, int commit)
{
	if (mp->keyfile) {
		fclose(mp->keyfile);
		mp->keyfile = 0;
	}
	if (mp->lockfile) {
		char	lockname[256];

		sprintf(lockname, "%s.lock", KEYFILE);
		fclose(mp->lockfile);
		mp->lockfile = 0;

		if (!commit) {
			unlink(lockname);
		} else
		if (rename(lockname, KEYFILE) < 0) {
			syslog(LOG_ERR, "rename %s: %m", lockname);
			return -1;
		}
	} else if (commit) {
		syslog(LOG_ERR,
			"internal error: commit but no lock file!");
		return -1;
	}
	return 0;
}

static int
skey_copyperms(int lockfd)
{
	struct stat	stb;

	if (stat(KEYFILE, &stb) < 0)
		return -1;
	if (fchmod(lockfd, stb.st_mode & 0666) < 0
	 || fchown(lockfd, stb.st_uid, stb.st_gid)  < 0)
		return -1;
	return 0;
}

static int
skey_getline(struct skey *mp, char *buffer, unsigned int size)
{
	char	*cp;

	buffer[0] = '\0';
	if (fgets(buffer, size, mp->keyfile) == NULL)
		return -1;
	strncpy(mp->buf, buffer, SKEY_BUFSIZE);
	mp->buf[SKEY_BUFSIZE-1] = '\0';
	rip(mp->buf);

	/* Empty line or comment */
	if (!(cp = strtok(mp->buf, " \t")) || *cp == '#')
		return -1;
	mp->logname  = cp;

	if (!(cp = strtok(NULL, " \t")))
		return -1;
	mp->n = atoi(cp);
	if ((mp->seed = strtok(NULL, " \t")) == NULL)
		return -1;
	if ((cp = strtok(NULL, " \t")) == NULL)
		return -1;
	if (atob8(mp->keybuf, cp) < 0)
		return -1;
	mp->key = mp->keybuf;

	return 0;
}

static void
skey_putline(struct skey *mp)
{
	time_t		now;
	struct tm	*tm;
	char		timestr[64], keystr[64];

	time(&now);
	tm = localtime(&now);
	strftime(timestr, sizeof(timestr), "%b %d,%Y %T", tm);
	btoa8(keystr, mp->key);
	fprintf(mp->lockfile, "%-10s %04d %-16s %-20s %-21s\n",
		mp->logname, mp->n, mp->seed, keystr, timestr);
}
