/*
 * skeysubr.c - subroutines for S/Key handling
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include "skey.h"

static void	docrunch(unsigned char result[8], void *, unsigned int);

static inline void
sevenbit(unsigned char *s)
{
	/* make sure there are only 7 bit code in the line */
	while (*s)
		*s++ &= 0x7f;
}

/* Crunch a key:
 * concatenate the seed and the password, run through hash and
 * collapse to 64 bits. This is defined as the user's starting key.
 */
int
keycrunch(unsigned char result[8], const char *seed, const char *passwd)
{
	unsigned int    buflen;
	char		*buf;

	/* Concatenate seed and password, and force to 7 bits */
	buflen = strlen(seed) + strlen(passwd);
	if ((buf = alloca(buflen + 1)) == NULL)
		return -1;
	strcpy(buf, seed);
	strcat(buf, passwd);
	sevenbit(buf);

	/* Crunch the key through MD[45] */
	docrunch(result, buf, buflen);
	return 0;
}

/* The one-way function f(). Takes 8 bytes and returns 8 bytes in place */
void
f(unsigned char x[8])
{
	docrunch(x, x, 8);
}

/* Strip trailing cr/lf from a line of text */
void
rip(char *buf)
{
	buf[strcspn(buf, "\r\n")] = '\0';
}

/*
 * Hash handling using MD5 (or some other hash algorithm, like maybe SHA-1):
 */
#ifdef USE_MD5
#include "md5.c"

static void
docrunch(unsigned char result[8], void *data, unsigned int len)
{
	MD5_CTX		ctxt;
	unsigned char	temp[16];
	unsigned int	j;

	MD5Init(&ctxt);
	MD5Update(&ctxt, (unsigned char *) data, len);
	MD5Final(temp, &ctxt);

	/* Fold to 64 bits */
	for (j = 0; j < 8; j++)
		result[j] = temp[j] ^ (temp+8)[j];
}
#endif

#if 0
static struct termios saved_ttymode;

static void
interrupt(sig)
     int             sig;
{
	tcsetattr(0, TCSANOW, &saved_ttymode);
	exit(1);
}

char           *
readpass(buf, n)
     char           *buf;
     int             n;
{
	struct termios  noecho_ttymode;
	void            (*oldsig) ();

	/* Save normal line editing modes */
	tcgetattr(0, &saved_ttymode);
	if ((oldsig = signal(SIGINT, SIG_IGN)) != SIG_IGN)
		signal(SIGINT, interrupt);

	/* Turn off echoing */
	tcgetattr(0, &noecho_ttymode);
	noecho_ttymode.c_lflag &= ~ECHO;
	tcsetattr(0, TCSANOW, &noecho_ttymode);
	fgets(buf, n, stdin);
	rip(buf);

	/* Restore previous tty modes */
	tcsetattr(0, TCSANOW, &saved_ttymode);
	if (oldsig != SIG_IGN)
		signal(SIGINT, oldsig);

	/*
	   after the secret key is taken from the keyboard, the line feed is
	   written to standard error instead of standard output.  That means that
	   anyone using the program from a terminal won't notice, but capturing
	   standard output will get the key words without a newline in front of
	   them. 
	 */
	fprintf(stderr, "\n");
	fflush(stderr);
	sevenbit(buf);

	return buf;
}
#endif
